@@ -1,27 +1,65 @@
+#![doc = include_str!("../README.md")]
+
 use anyhow::{anyhow, Context, Error, Result};
 use libloading::{Library, Symbol};
 use once_cell::unsync::OnceCell;
 use regex::{Regex, RegexBuilder};
-use serde_derive::{Deserialize, Serialize};
+use serde::{Deserialize, Deserializer, Serialize};
 use std::collections::HashMap;
+use std::ffi::{OsStr, OsString};
 use std::io::BufReader;
 use std::ops::Range;
 use std::path::{Path, PathBuf};
 use std::process::Command;
 use std::sync::Mutex;
 use std::time::SystemTime;
-use std::{fs, mem};
-use tree_sitter::{Language, QueryError};
+use std::{env, fs, mem};
+use tree_sitter::{Language, QueryError, QueryErrorKind};
 use tree_sitter_highlight::HighlightConfiguration;
 use tree_sitter_tags::{Error as TagsError, TagsConfiguration};
+use which::which;
+
+pub const EMSCRIPTEN_TAG: &'static str = concat!("emscripten/emsdk:", env!("EMSCRIPTEN_VERSION"));
 
 #[derive(Default, Deserialize, Serialize)]
 pub struct Config {
     #[serde(default)]
-    #[serde(rename = "parser-directories")]
+    #[serde(
+        rename = "parser-directories",
+        deserialize_with = "deserialize_parser_directories"
+    )]
     pub parser_directories: Vec<PathBuf>,
 }
 
+// Replace `~` or `$HOME` with home path string.
+// (While paths like "~/.tree-sitter/config.json" can be deserialized,
+// they're not valid path for I/O modules.)
+fn deserialize_parser_directories<'de, D>(deserializer: D) -> Result<Vec<PathBuf>, D::Error>
+where
+    D: Deserializer<'de>,
+{
+    let paths = Vec::<PathBuf>::deserialize(deserializer)?;
+    let home = match dirs::home_dir() {
+        Some(home) => home,
+        None => return Ok(paths),
+    };
+    let standardized = paths
+        .into_iter()
+        .map(|path| standardize_path(path, &home))
+        .collect();
+    Ok(standardized)
+}
+
+fn standardize_path(path: PathBuf, home: &Path) -> PathBuf {
+    if let Ok(p) = path.strip_prefix("~") {
+        return home.join(p);
+    }
+    if let Ok(p) = path.strip_prefix("$HOME") {
+        return home.join(p);
+    }
+    path
+}
+
 impl Config {
     pub fn initial() -> Config {
         let home_dir = dirs::home_dir().expect("Cannot determine home directory");
@@ -36,12 +74,12 @@ impl Config {
 }
 
 #[cfg(unix)]
-const DYLIB_EXTENSION: &'static str = "so";
+const DYLIB_EXTENSION: &str = "so";
 
 #[cfg(windows)]
 const DYLIB_EXTENSION: &'static str = "dll";
 
-const BUILD_TARGET: &'static str = env!("BUILD_TARGET");
+const BUILD_TARGET: &str = env!("BUILD_TARGET");
 
 pub struct LanguageConfiguration<'a> {
     pub scope: Option<String>,
@@ -54,6 +92,7 @@ pub struct LanguageConfiguration<'a> {
     pub injections_filenames: Option<Vec<String>>,
     pub locals_filenames: Option<Vec<String>>,
     pub tags_filenames: Option<Vec<String>>,
+    pub language_name: String,
     language_id: usize,
     highlight_config: OnceCell<Option<HighlightConfiguration>>,
     tags_config: OnceCell<Option<TagsConfiguration>>,
@@ -66,8 +105,13 @@ pub struct Loader {
     languages_by_id: Vec<(PathBuf, OnceCell<Language>)>,
     language_configurations: Vec<LanguageConfiguration<'static>>,
     language_configuration_ids_by_file_type: HashMap<String, Vec<usize>>,
+    language_configuration_in_current_path: Option<usize>,
     highlight_names: Box<Mutex<Vec<String>>>,
     use_all_highlight_names: bool,
+    debug_build: bool,
+
+    #[cfg(feature = "wasm")]
+    wasm_store: Mutex<Option<tree_sitter::WasmStore>>,
 }
 
 unsafe impl Send for Loader {}
@@ -75,9 +119,13 @@ unsafe impl Sync for Loader {}
 
 impl Loader {
     pub fn new() -> Result<Self> {
-        let parser_lib_path = dirs::cache_dir()
-            .ok_or(anyhow!("Cannot determine cache directory"))?
-            .join("tree-sitter/lib");
+        let parser_lib_path = match env::var("TREE_SITTER_LIBDIR") {
+            Ok(path) => PathBuf::from(path),
+            _ => dirs::cache_dir()
+                .ok_or(anyhow!("Cannot determine cache directory"))?
+                .join("tree-sitter")
+                .join("lib"),
+        };
         Ok(Self::with_parser_lib_path(parser_lib_path))
     }
 
@@ -87,12 +135,17 @@ impl Loader {
             languages_by_id: Vec::new(),
             language_configurations: Vec::new(),
             language_configuration_ids_by_file_type: HashMap::new(),
+            language_configuration_in_current_path: None,
             highlight_names: Box::new(Mutex::new(Vec::new())),
             use_all_highlight_names: true,
+            debug_build: false,
+
+            #[cfg(feature = "wasm")]
+            wasm_store: Default::default(),
         }
     }
 
-    pub fn configure_highlights(&mut self, names: &Vec<String>) {
+    pub fn configure_highlights(&mut self, names: &[String]) {
         self.use_all_highlight_names = false;
         let mut highlights = self.highlight_names.lock().unwrap();
         highlights.clear();
@@ -108,8 +161,7 @@ impl Loader {
             eprintln!("Warning: You have not configured any parser directories!");
             eprintln!("Please run `tree-sitter init-config` and edit the resulting");
             eprintln!("configuration file to indicate where we should look for");
-            eprintln!("language grammars.");
-            eprintln!("");
+            eprintln!("language grammars.\n");
         }
         for parser_container_dir in &config.parser_directories {
             if let Ok(entries) = fs::read_dir(parser_container_dir) {
@@ -119,6 +171,7 @@ impl Loader {
                         if parser_dir_name.starts_with("tree-sitter-") {
                             self.find_language_configurations_at_path(
                                 &parser_container_dir.join(parser_dir_name),
+                                false,
                             )
                             .ok();
                         }
@@ -130,7 +183,7 @@ impl Loader {
     }
 
     pub fn languages_at_path(&mut self, path: &Path) -> Result<Vec<Language>> {
-        if let Ok(configurations) = self.find_language_configurations_at_path(path) {
+        if let Ok(configurations) = self.find_language_configurations_at_path(path, true) {
             let mut language_ids = configurations
                 .iter()
                 .map(|c| c.language_id)
@@ -276,8 +329,6 @@ impl Loader {
 
     pub fn load_language_at_path(&self, src_path: &Path, header_path: &Path) -> Result<Language> {
         let grammar_path = src_path.join("grammar.json");
-        let parser_path = src_path.join("parser.c");
-        let mut scanner_path = src_path.join("scanner.c");
 
         #[derive(Deserialize)]
         struct GrammarJSON {
@@ -288,114 +339,291 @@ impl Loader {
         let grammar_json: GrammarJSON = serde_json::from_reader(BufReader::new(&mut grammar_file))
             .with_context(|| "Failed to parse grammar.json")?;
 
-        let scanner_path = if scanner_path.exists() {
-            Some(scanner_path)
-        } else {
-            scanner_path.set_extension("cc");
-            if scanner_path.exists() {
-                Some(scanner_path)
-            } else {
-                None
-            }
-        };
-
-        self.load_language_from_sources(
-            &grammar_json.name,
-            &header_path,
-            &parser_path,
-            &scanner_path,
-        )
+        self.load_language_at_path_with_name(src_path, &header_path, &grammar_json.name)
     }
 
-    pub fn load_language_from_sources(
+    pub fn load_language_at_path_with_name(
         &self,
+        src_path: &Path,
+        header_path: &Path,
         name: &str,
+    ) -> Result<Language> {
+        let mut lib_name = name.to_string();
+        let language_fn_name = format!("tree_sitter_{}", replace_dashes_with_underscores(name));
+        if self.debug_build {
+            lib_name.push_str(".debug._");
+        }
+
+        fs::create_dir_all(&self.parser_lib_path)?;
+
+        let mut library_path = self.parser_lib_path.join(lib_name);
+        library_path.set_extension(DYLIB_EXTENSION);
+
+        let parser_path = src_path.join("parser.c");
+        let scanner_path = self.get_scanner_path(&src_path);
+
+        #[cfg(feature = "wasm")]
+        if self.wasm_store.lock().unwrap().is_some() {
+            library_path.set_extension("wasm");
+        }
+
+        let recompile = needs_recompile(&library_path, &parser_path, scanner_path.as_deref())
+            .with_context(|| "Failed to compare source and binary timestamps")?;
+
+        #[cfg(feature = "wasm")]
+        if let Some(wasm_store) = self.wasm_store.lock().unwrap().as_mut() {
+            if recompile {
+                self.compile_parser_to_wasm(
+                    name,
+                    src_path,
+                    scanner_path
+                        .as_ref()
+                        .and_then(|p| p.strip_prefix(&src_path).ok()),
+                    &library_path,
+                    false,
+                )?;
+            }
+
+            let wasm_bytes = fs::read(&library_path)?;
+            return Ok(wasm_store.load_language(name, &wasm_bytes)?);
+        }
+
+        {
+            if recompile {
+                self.compile_parser_to_dylib(
+                    header_path,
+                    &parser_path,
+                    &scanner_path,
+                    &library_path,
+                )?;
+            }
+
+            let library = unsafe { Library::new(&library_path) }
+                .with_context(|| format!("Error opening dynamic library {:?}", &library_path))?;
+            let language = unsafe {
+                let language_fn: Symbol<unsafe extern "C" fn() -> Language> = library
+                    .get(language_fn_name.as_bytes())
+                    .with_context(|| format!("Failed to load symbol {}", language_fn_name))?;
+                language_fn()
+            };
+            mem::forget(library);
+            return Ok(language);
+        }
+    }
+
+    fn compile_parser_to_dylib(
+        &self,
         header_path: &Path,
         parser_path: &Path,
         scanner_path: &Option<PathBuf>,
-    ) -> Result<Language> {
-        let mut library_path = self.parser_lib_path.join(name);
-        library_path.set_extension(DYLIB_EXTENSION);
+        library_path: &PathBuf,
+    ) -> Result<(), Error> {
+        let mut config = cc::Build::new();
+        config
+            .cpp(true)
+            .opt_level(2)
+            .cargo_metadata(false)
+            .target(BUILD_TARGET)
+            .host(BUILD_TARGET)
+            .flag_if_supported("-Werror=implicit-function-declaration");
+        let compiler = config.get_compiler();
+        let mut command = Command::new(compiler.path());
+        for (key, value) in compiler.env() {
+            command.env(key, value);
+        }
 
-        let recompile = needs_recompile(&library_path, &parser_path, &scanner_path)
-            .with_context(|| "Failed to compare source and binary timestamps")?;
+        if compiler.is_like_msvc() {
+            command.args(&["/nologo", "/LD", "/I"]).arg(header_path);
+            if self.debug_build {
+                command.arg("/Od");
+            } else {
+                command.arg("/O2");
+            }
+            command.arg(parser_path);
+            if let Some(scanner_path) = scanner_path.as_ref() {
+                command.arg(scanner_path);
+            }
+            command
+                .arg("/link")
+                .arg(format!("/out:{}", library_path.to_str().unwrap()));
+        } else {
+            command
+                .arg("-shared")
+                .arg("-fno-exceptions")
+                .arg("-g")
+                .arg("-I")
+                .arg(header_path)
+                .arg("-o")
+                .arg(&library_path);
 
-        if recompile {
-            fs::create_dir_all(&self.parser_lib_path)?;
-            let mut config = cc::Build::new();
-            config
-                .cpp(true)
-                .opt_level(2)
-                .cargo_metadata(false)
-                .target(BUILD_TARGET)
-                .host(BUILD_TARGET);
-            let compiler = config.get_compiler();
-            let mut command = Command::new(compiler.path());
-            for (key, value) in compiler.env() {
-                command.env(key, value);
+            if !cfg!(windows) {
+                command.arg("-fPIC");
             }
 
-            if cfg!(windows) {
-                command
-                    .args(&["/nologo", "/LD", "/I"])
-                    .arg(header_path)
-                    .arg("/Od")
-                    .arg(parser_path);
-                if let Some(scanner_path) = scanner_path.as_ref() {
+            if self.debug_build {
+                command.arg("-O0");
+            } else {
+                command.arg("-O2");
+            }
+
+            // For conditional compilation of external scanner code when
+            // used internally by `tree-siteer parse` and other sub commands.
+            command.arg("-DTREE_SITTER_INTERNAL_BUILD");
+
+            if let Some(scanner_path) = scanner_path.as_ref() {
+                if scanner_path.extension() == Some("c".as_ref()) {
+                    command.arg("-xc").arg("-std=c99").arg(scanner_path);
+                } else {
                     command.arg(scanner_path);
                 }
-                command
-                    .arg("/link")
-                    .arg(format!("/out:{}", library_path.to_str().unwrap()));
-            } else {
-                command
-                    .arg("-shared")
-                    .arg("-fPIC")
-                    .arg("-fno-exceptions")
-                    .arg("-g")
-                    .arg("-I")
-                    .arg(header_path)
-                    .arg("-o")
-                    .arg(&library_path)
-                    .arg("-O2");
-                if let Some(scanner_path) = scanner_path.as_ref() {
-                    if scanner_path.extension() == Some("c".as_ref()) {
-                        command.arg("-xc").arg("-std=c99").arg(scanner_path);
-                    } else {
-                        command.arg(scanner_path);
+            }
+            command.arg("-xc").arg(parser_path);
+        }
+
+        let output = command
+            .output()
+            .with_context(|| "Failed to execute C compiler")?;
+        if !output.status.success() {
+            return Err(anyhow!(
+                "Parser compilation failed.\nStdout: {}\nStderr: {}",
+                String::from_utf8_lossy(&output.stdout),
+                String::from_utf8_lossy(&output.stderr)
+            ));
+        }
+
+        #[cfg(any(target_os = "macos", target_os = "linux"))]
+        if scanner_path.is_some() {
+            let command = Command::new("nm")
+                .arg("-W")
+                .arg("-U")
+                .arg(&library_path)
+                .output();
+            if let Ok(output) = command {
+                if output.status.success() {
+                    let mut found_non_static = false;
+                    for line in String::from_utf8_lossy(&output.stdout).lines() {
+                        if line.contains(" T ") && !line.contains("tree_sitter_") {
+                            if let Some(function_name) =
+                                line.split_whitespace().collect::<Vec<_>>().get(2)
+                            {
+                                if !found_non_static {
+                                    found_non_static = true;
+                                    eprintln!("Warning: Found non-static non-tree-sitter functions in external scannner");
+                                }
+                                eprintln!("  `{function_name}`");
+                            }
+                        }
+                    }
+                    if found_non_static {
+                        eprintln!("Consider making these functions static, they can cause conflicts when another tree-sitter project uses the same function name");
                     }
                 }
-                command.arg("-xc").arg(parser_path);
-            }
-
-            let output = command
-                .output()
-                .with_context(|| "Failed to execute C compiler")?;
-            if !output.status.success() {
-                return Err(anyhow!(
-                    "Parser compilation failed.\nStdout: {}\nStderr: {}",
-                    String::from_utf8_lossy(&output.stdout),
-                    String::from_utf8_lossy(&output.stderr)
-                ));
             }
         }
 
-        let library = unsafe { Library::new(&library_path) }
-            .with_context(|| format!("Error opening dynamic library {:?}", &library_path))?;
-        let language_fn_name = format!("tree_sitter_{}", replace_dashes_with_underscores(name));
-        let language = unsafe {
-            let language_fn: Symbol<unsafe extern "C" fn() -> Language> = library
-                .get(language_fn_name.as_bytes())
-                .with_context(|| format!("Failed to load symbol {}", language_fn_name))?;
-            language_fn()
-        };
-        mem::forget(library);
-        Ok(language)
+        Ok(())
+    }
+
+    pub fn compile_parser_to_wasm(
+        &self,
+        language_name: &str,
+        src_path: &Path,
+        scanner_filename: Option<&Path>,
+        output_path: &PathBuf,
+        force_docker: bool,
+    ) -> Result<(), Error> {
+        let emcc_bin = if cfg!(windows) { "emcc.bat" } else { "emcc" };
+        let emcc_path = which(emcc_bin)
+            .ok()
+            .and_then(|p| Command::new(&p).output().and(Ok(p)).ok());
+
+        let mut command;
+        if emcc_path.is_some() && !force_docker {
+            command = Command::new(emcc_path.unwrap());
+            command.current_dir(&src_path);
+        } else if Command::new("docker").output().is_ok() {
+            command = Command::new("docker");
+            command.args(&["run", "--rm"]);
+
+            // Mount the parser directory as a volume
+            command.args(&["--workdir", "/src"]);
+
+            let mut volume_string = OsString::from(&src_path);
+            volume_string.push(":/src:Z");
+            command.args(&[OsStr::new("--volume"), &volume_string]);
+
+            // Get the current user id so that files created in the docker container will have
+            // the same owner.
+            if cfg!(unix) {
+                let user_id_output = Command::new("id")
+                    .arg("-u")
+                    .output()
+                    .with_context(|| "Failed to get get current user id")?;
+                let user_id = String::from_utf8_lossy(&user_id_output.stdout);
+                let user_id = user_id.trim();
+                command.args(&["--user", user_id]);
+            }
+
+            // Run `emcc` in a container using the `emscripten-slim` image
+            command.args(&[EMSCRIPTEN_TAG, "emcc"]);
+        } else {
+            return Err(anyhow!(
+                "You must have either emcc or docker on your PATH to run this command"
+            ));
+        }
+
+        let output_name = "output.wasm";
+
+        command.args(&[
+            "-o",
+            output_name,
+            "-Os",
+            "-s",
+            "WASM=1",
+            "-s",
+            "SIDE_MODULE=1",
+            "-s",
+            "TOTAL_MEMORY=33554432",
+            "-s",
+            "NODEJS_CATCH_EXIT=0",
+            "-s",
+            &format!("EXPORTED_FUNCTIONS=[\"_tree_sitter_{}\"]", language_name),
+            "-fno-exceptions",
+            "-fvisibility=hidden",
+            "-I",
+            ".",
+        ]);
+
+        if let Some(scanner_filename) = scanner_filename {
+            if scanner_filename
+                .extension()
+                .and_then(|ext| ext.to_str())
+                .map_or(false, |ext| ["cc", "cpp"].contains(&ext))
+            {
+                command.arg("-xc++");
+            }
+            command.arg(&scanner_filename);
+        }
+
+        command.arg("parser.c");
+
+        let output = command.output().context("Failed to run emcc command")?;
+        if !output.status.success() {
+            return Err(anyhow!(
+                "emcc command failed - {}",
+                String::from_utf8_lossy(&output.stderr)
+            ));
+        }
+
+        fs::rename(&src_path.join(output_name), &output_path)
+            .context("failed to rename wasm output file")?;
+        Ok(())
     }
 
     pub fn highlight_config_for_injection_string<'a>(
         &'a self,
         string: &str,
+        apply_all_captures: bool,
     ) -> Option<&'a HighlightConfiguration> {
         match self.language_configuration_for_injection_string(string) {
             Err(e) => {
@@ -406,38 +634,36 @@ impl Loader {
                 None
             }
             Ok(None) => None,
-            Ok(Some((language, configuration))) => match configuration.highlight_config(language) {
-                Err(e) => {
-                    eprintln!(
-                        "Failed to load property sheet for injection string '{}': {}",
-                        string, e
-                    );
-                    None
+            Ok(Some((language, configuration))) => {
+                match configuration.highlight_config(language, apply_all_captures, None) {
+                    Err(e) => {
+                        eprintln!(
+                            "Failed to load property sheet for injection string '{}': {}",
+                            string, e
+                        );
+                        None
+                    }
+                    Ok(None) => None,
+                    Ok(Some(config)) => Some(config),
                 }
-                Ok(None) => None,
-                Ok(Some(config)) => Some(config),
-            },
+            }
         }
     }
 
-    pub fn find_language_configurations_at_path<'a>(
-        &'a mut self,
+    pub fn find_language_configurations_at_path(
+        &mut self,
         parser_path: &Path,
+        set_current_path_config: bool,
     ) -> Result<&[LanguageConfiguration]> {
-        #[derive(Deserialize)]
+        #[derive(Default, Deserialize)]
         #[serde(untagged)]
         enum PathsJSON {
+            #[default]
             Empty,
             Single(String),
             Multiple(Vec<String>),
         }
 
-        impl Default for PathsJSON {
-            fn default() -> Self {
-                PathsJSON::Empty
-            }
-        }
-
         impl PathsJSON {
             fn into_vec(self) -> Option<Vec<String>> {
                 match self {
@@ -478,6 +704,11 @@ impl Loader {
             tree_sitter: Vec<LanguageConfigurationJSON>,
         }
 
+        #[derive(Deserialize)]
+        struct GrammarJSON {
+            name: String,
+        }
+
         let initial_language_configuration_count = self.language_configurations.len();
 
         if let Ok(package_json_contents) = fs::read_to_string(&parser_path.join("package.json")) {
@@ -489,6 +720,13 @@ impl Loader {
                     // the package.json, but defaults to the directory containing the package.json.
                     let language_path = parser_path.join(config_json.path);
 
+                    let grammar_path = language_path.join("src").join("grammar.json");
+                    let mut grammar_file = fs::File::open(grammar_path)
+                        .with_context(|| "Failed to read grammar.json")?;
+                    let grammar_json: GrammarJSON =
+                        serde_json::from_reader(BufReader::new(&mut grammar_file))
+                            .with_context(|| "Failed to parse grammar.json")?;
+
                     // Determine if a previous language configuration in this package.json file
                     // already uses the same language.
                     let mut language_id = None;
@@ -508,6 +746,7 @@ impl Loader {
 
                     let configuration = LanguageConfiguration {
                         root_path: parser_path.to_path_buf(),
+                        language_name: grammar_json.name.clone(),
                         scope: config_json.scope,
                         language_id,
                         file_types: config_json.file_types.unwrap_or(Vec::new()),
@@ -520,19 +759,26 @@ impl Loader {
                         highlights_filenames: config_json.highlights.into_vec(),
                         highlight_config: OnceCell::new(),
                         tags_config: OnceCell::new(),
-                        highlight_names: &*self.highlight_names,
+                        highlight_names: &self.highlight_names,
                         use_all_highlight_names: self.use_all_highlight_names,
                     };
 
                     for file_type in &configuration.file_types {
                         self.language_configuration_ids_by_file_type
                             .entry(file_type.to_string())
-                            .or_insert(Vec::new())
+                            .or_default()
                             .push(self.language_configurations.len());
                     }
 
                     self.language_configurations
                         .push(unsafe { mem::transmute(configuration) });
+
+                    if set_current_path_config
+                        && self.language_configuration_in_current_path.is_none()
+                    {
+                        self.language_configuration_in_current_path =
+                            Some(self.language_configurations.len() - 1);
+                    }
                 }
             }
         }
@@ -540,8 +786,15 @@ impl Loader {
         if self.language_configurations.len() == initial_language_configuration_count
             && parser_path.join("src").join("grammar.json").exists()
         {
+            let grammar_path = parser_path.join("src").join("grammar.json");
+            let mut grammar_file =
+                fs::File::open(grammar_path).with_context(|| "Failed to read grammar.json")?;
+            let grammar_json: GrammarJSON =
+                serde_json::from_reader(BufReader::new(&mut grammar_file))
+                    .with_context(|| "Failed to parse grammar.json")?;
             let configuration = LanguageConfiguration {
                 root_path: parser_path.to_owned(),
+                language_name: grammar_json.name,
                 language_id: self.languages_by_id.len(),
                 file_types: Vec::new(),
                 scope: None,
@@ -554,7 +807,7 @@ impl Loader {
                 tags_filenames: None,
                 highlight_config: OnceCell::new(),
                 tags_config: OnceCell::new(),
-                highlight_names: &*self.highlight_names,
+                highlight_names: &self.highlight_names,
                 use_all_highlight_names: self.use_all_highlight_names,
             };
             self.language_configurations
@@ -579,11 +832,11 @@ impl Loader {
         if let Some(scope) = scope {
             if let Some(config) = self
                 .language_configuration_for_scope(scope)
-                .with_context(|| format!("Failed to load language for scope '{}'", scope))?
+                .with_context(|| format!("Failed to load language for scope '{scope}'"))?
             {
                 Ok(config.0)
             } else {
-                return Err(anyhow!("Unknown scope '{}'", scope));
+                Err(anyhow!("Unknown scope '{scope}'"))
             }
         } else if let Some((lang, _)) = self
             .language_configuration_for_file_name(path)
@@ -595,8 +848,10 @@ impl Loader {
             })?
         {
             Ok(lang)
+        } else if let Some(id) = self.language_configuration_in_current_path {
+            Ok(self.language_for_id(self.language_configurations[id].language_id)?)
         } else if let Some(lang) = self
-            .languages_at_path(&current_dir)
+            .languages_at_path(current_dir)
             .with_context(|| "Failed to load language in current directory")?
             .first()
             .cloned()
@@ -606,62 +861,136 @@ impl Loader {
             Err(anyhow!("No language found"))
         }
     }
+
+    pub fn use_debug_build(&mut self, flag: bool) {
+        self.debug_build = flag;
+    }
+
+    #[cfg(feature = "wasm")]
+    pub fn use_wasm(&mut self, engine: tree_sitter::wasmtime::Engine) {
+        *self.wasm_store.lock().unwrap() = Some(tree_sitter::WasmStore::new(engine).unwrap())
+    }
+
+    pub fn get_scanner_path(&self, src_path: &Path) -> Option<PathBuf> {
+        let mut path = src_path.join("scanner.c");
+        for extension in ["c", "cc", "cpp"] {
+            path.set_extension(extension);
+            if path.exists() {
+                return Some(path);
+            }
+        }
+        None
+    }
 }
 
 impl<'a> LanguageConfiguration<'a> {
-    pub fn highlight_config(&self, language: Language) -> Result<Option<&HighlightConfiguration>> {
+    pub fn highlight_config(
+        &self,
+        language: Language,
+        apply_all_captures: bool,
+        paths: Option<&[String]>,
+    ) -> Result<Option<&HighlightConfiguration>> {
+        let (highlights_filenames, injections_filenames, locals_filenames) = match paths {
+            Some(paths) => (
+                Some(
+                    paths
+                        .iter()
+                        .filter(|p| p.ends_with("highlights.scm"))
+                        .cloned()
+                        .collect::<Vec<_>>(),
+                ),
+                Some(
+                    paths
+                        .iter()
+                        .filter(|p| p.ends_with("tags.scm"))
+                        .cloned()
+                        .collect::<Vec<_>>(),
+                ),
+                Some(
+                    paths
+                        .iter()
+                        .filter(|p| p.ends_with("locals.scm"))
+                        .cloned()
+                        .collect::<Vec<_>>(),
+                ),
+            ),
+            None => (None, None, None),
+        };
         return self
             .highlight_config
             .get_or_try_init(|| {
-                let (highlights_query, highlight_ranges) =
-                    self.read_queries(&self.highlights_filenames, "highlights.scm")?;
-                let (injections_query, injection_ranges) =
-                    self.read_queries(&self.injections_filenames, "injections.scm")?;
-                let (locals_query, locals_ranges) =
-                    self.read_queries(&self.locals_filenames, "locals.scm")?;
+                let (highlights_query, highlight_ranges) = self.read_queries(
+                    if highlights_filenames.is_some() {
+                        highlights_filenames.as_deref()
+                    } else {
+                        self.highlights_filenames.as_deref()
+                    },
+                    "highlights.scm",
+                )?;
+                let (injections_query, injection_ranges) = self.read_queries(
+                    if injections_filenames.is_some() {
+                        injections_filenames.as_deref()
+                    } else {
+                        self.injections_filenames.as_deref()
+                    },
+                    "injections.scm",
+                )?;
+                let (locals_query, locals_ranges) = self.read_queries(
+                    if locals_filenames.is_some() {
+                        locals_filenames.as_deref()
+                    } else {
+                        self.locals_filenames.as_deref()
+                    },
+                    "locals.scm",
+                )?;
 
                 if highlights_query.is_empty() {
                     Ok(None)
                 } else {
                     let mut result = HighlightConfiguration::new(
                         language,
+                        &self.language_name,
                         &highlights_query,
                         &injections_query,
                         &locals_query,
+                        apply_all_captures,
                     )
-                    .map_err(|error| {
-                        if error.offset < injections_query.len() {
-                            Self::include_path_in_query_error(
-                                error,
-                                &injection_ranges,
-                                &injections_query,
-                                0,
-                            )
-                        } else if error.offset < injections_query.len() + locals_query.len() {
-                            Self::include_path_in_query_error(
-                                error,
-                                &locals_ranges,
-                                &locals_query,
-                                injections_query.len(),
-                            )
-                        } else {
-                            Self::include_path_in_query_error(
-                                error,
-                                &highlight_ranges,
-                                &highlights_query,
-                                injections_query.len() + locals_query.len(),
-                            )
+                    .map_err(|error| match error.kind {
+                        QueryErrorKind::Language => Error::from(error),
+                        _ => {
+                            if error.offset < injections_query.len() {
+                                Self::include_path_in_query_error(
+                                    error,
+                                    &injection_ranges,
+                                    &injections_query,
+                                    0,
+                                )
+                            } else if error.offset < injections_query.len() + locals_query.len() {
+                                Self::include_path_in_query_error(
+                                    error,
+                                    &locals_ranges,
+                                    &locals_query,
+                                    injections_query.len(),
+                                )
+                            } else {
+                                Self::include_path_in_query_error(
+                                    error,
+                                    &highlight_ranges,
+                                    &highlights_query,
+                                    injections_query.len() + locals_query.len(),
+                                )
+                            }
                         }
                     })?;
                     let mut all_highlight_names = self.highlight_names.lock().unwrap();
                     if self.use_all_highlight_names {
                         for capture_name in result.query.capture_names() {
-                            if !all_highlight_names.contains(capture_name) {
-                                all_highlight_names.push(capture_name.clone());
+                            if !all_highlight_names.iter().any(|x| x == capture_name) {
+                                all_highlight_names.push(capture_name.to_string());
                             }
                         }
                     }
-                    result.configure(&all_highlight_names.as_slice());
+                    result.configure(all_highlight_names.as_slice());
                     Ok(Some(result))
                 }
             })
@@ -672,9 +1001,9 @@ impl<'a> LanguageConfiguration<'a> {
         self.tags_config
             .get_or_try_init(|| {
                 let (tags_query, tags_ranges) =
-                    self.read_queries(&self.tags_filenames, "tags.scm")?;
+                    self.read_queries(self.tags_filenames.as_deref(), "tags.scm")?;
                 let (locals_query, locals_ranges) =
-                    self.read_queries(&self.locals_filenames, "locals.scm")?;
+                    self.read_queries(self.locals_filenames.as_deref(), "locals.scm")?;
                 if tags_query.is_empty() {
                     Ok(None)
                 } else {
@@ -697,7 +1026,6 @@ impl<'a> LanguageConfiguration<'a> {
                                         locals_query.len(),
                                     )
                                 }
-                                .into()
                             } else {
                                 error.into()
                             }
@@ -707,9 +1035,9 @@ impl<'a> LanguageConfiguration<'a> {
             .map(Option::as_ref)
     }
 
-    fn include_path_in_query_error<'b>(
+    fn include_path_in_query_error(
         mut error: QueryError,
-        ranges: &'b Vec<(String, Range<usize>)>,
+        ranges: &[(String, Range<usize>)],
         source: &str,
         start_offset: usize,
     ) -> Error {
@@ -717,7 +1045,7 @@ impl<'a> LanguageConfiguration<'a> {
         let (path, range) = ranges
             .iter()
             .find(|(_, range)| range.contains(&offset_within_section))
-            .unwrap();
+            .unwrap_or(ranges.last().unwrap());
         error.offset = offset_within_section - range.start;
         error.row = source[range.start..offset_within_section]
             .chars()
@@ -728,12 +1056,12 @@ impl<'a> LanguageConfiguration<'a> {
 
     fn read_queries(
         &self,
-        paths: &Option<Vec<String>>,
+        paths: Option<&[String]>,
         default_path: &str,
     ) -> Result<(String, Vec<(String, Range<usize>)>)> {
         let mut query = String::new();
         let mut path_ranges = Vec::new();
-        if let Some(paths) = paths.as_ref() {
+        if let Some(paths) = paths {
             for path in paths {
                 let abs_path = self.root_path.join(path);
                 let prev_query_len = query.len();
@@ -758,7 +1086,7 @@ impl<'a> LanguageConfiguration<'a> {
 fn needs_recompile(
     lib_path: &Path,
     parser_c_path: &Path,
-    scanner_path: &Option<PathBuf>,
+    scanner_path: Option<&Path>,
 ) -> Result<bool> {
     if !lib_path.exists() {
         return Ok(true);
