@@ -20,10 +20,11 @@ use self::rules::AliasMap;
 use anyhow::{anyhow, Context, Result};
 use lazy_static::lazy_static;
 use regex::{Regex, RegexBuilder};
-use std::fs;
+use semver::Version;
 use std::io::Write;
 use std::path::{Path, PathBuf};
 use std::process::{Command, Stdio};
+use std::{env, fs};
 
 lazy_static! {
     static ref JSON_COMMENT_REGEX: Regex = RegexBuilder::new("^\\s*//.*")
@@ -43,25 +44,27 @@ pub fn generate_parser_in_directory(
     abi_version: usize,
     generate_bindings: bool,
     report_symbol_name: Option<&str>,
+    js_runtime: Option<&str>,
 ) -> Result<()> {
     let src_path = repo_path.join("src");
     let header_path = src_path.join("tree_sitter");
 
+    // Read the grammar.json.
+    let grammar_json = match grammar_path {
+        Some(path) => load_grammar_file(path.as_ref(), js_runtime)?,
+        None => {
+            let grammar_js_path = grammar_path.map_or(repo_path.join("grammar.js"), |s| s.into());
+            load_grammar_file(&grammar_js_path, js_runtime)?
+        }
+    };
+
     // Ensure that the output directories exist.
     fs::create_dir_all(&src_path)?;
     fs::create_dir_all(&header_path)?;
 
-    // Read the grammar.json.
-    let grammar_json;
-    match grammar_path {
-        Some(path) => {
-            grammar_json = load_grammar_file(path.as_ref())?;
-        }
-        None => {
-            let grammar_js_path = grammar_path.map_or(repo_path.join("grammar.js"), |s| s.into());
-            grammar_json = load_grammar_file(&grammar_js_path)?;
-            fs::write(&src_path.join("grammar.json"), &grammar_json)?;
-        }
+    if grammar_path.is_none() {
+        fs::write(&src_path.join("grammar.json"), &grammar_json)
+            .with_context(|| format!("Failed to write grammar.json to {:?}", src_path))?;
     }
 
     // Parse and preprocess the grammar.
@@ -154,10 +157,18 @@ fn generate_parser_for_grammar_with_opts(
     })
 }
 
-pub fn load_grammar_file(grammar_path: &Path) -> Result<String> {
+pub fn load_grammar_file(grammar_path: &Path, js_runtime: Option<&str>) -> Result<String> {
+    if grammar_path.is_dir() {
+        return Err(anyhow!(
+            "Path to a grammar file with `.js` or `.json` extension is required"
+        ));
+    }
     match grammar_path.extension().and_then(|e| e.to_str()) {
-        Some("js") => Ok(load_js_grammar_file(grammar_path)?),
-        Some("json") => Ok(fs::read_to_string(grammar_path)?),
+        Some("js") => Ok(load_js_grammar_file(grammar_path, js_runtime)
+            .with_context(|| "Failed to load grammar.js")?),
+        Some("json") => {
+            Ok(fs::read_to_string(grammar_path).with_context(|| "Failed to load grammar.json")?)
+        }
         _ => Err(anyhow!(
             "Unknown grammar file extension: {:?}",
             grammar_path
@@ -165,34 +176,47 @@ pub fn load_grammar_file(grammar_path: &Path) -> Result<String> {
     }
 }
 
-fn load_js_grammar_file(grammar_path: &Path) -> Result<String> {
+fn load_js_grammar_file(grammar_path: &Path, js_runtime: Option<&str>) -> Result<String> {
     let grammar_path = fs::canonicalize(grammar_path)?;
-    let mut node_process = Command::new("node")
+
+    let js_runtime = js_runtime.unwrap_or("node");
+
+    let mut node_process = Command::new(js_runtime)
         .env("TREE_SITTER_GRAMMAR_PATH", grammar_path)
         .stdin(Stdio::piped())
         .stdout(Stdio::piped())
         .spawn()
-        .expect("Failed to run `node`");
+        .with_context(|| format!("Failed to run `{js_runtime}`"))?;
 
     let mut node_stdin = node_process
         .stdin
         .take()
-        .expect("Failed to open stdin for node");
+        .with_context(|| "Failed to open stdin for node")?;
+    let cli_version = Version::parse(env!("CARGO_PKG_VERSION"))
+        .with_context(|| "Could not parse this package's version as semver.")?;
+    write!(
+        node_stdin,
+        "global.TREE_SITTER_CLI_VERSION_MAJOR = {};
+        global.TREE_SITTER_CLI_VERSION_MINOR = {};
+        global.TREE_SITTER_CLI_VERSION_PATCH = {};",
+        cli_version.major, cli_version.minor, cli_version.patch,
+    )
+    .with_context(|| "Failed to write tree-sitter version to node's stdin")?;
     let javascript_code = include_bytes!("./dsl.js");
     node_stdin
         .write(javascript_code)
-        .expect("Failed to write to node's stdin");
+        .with_context(|| "Failed to write grammar dsl to node's stdin")?;
     drop(node_stdin);
     let output = node_process
         .wait_with_output()
-        .expect("Failed to read output from node");
+        .with_context(|| "Failed to read output from node")?;
     match output.status.code() {
         None => panic!("Node process was killed"),
         Some(0) => {}
         Some(code) => return Err(anyhow!("Node process exited with status {}", code)),
     }
-
-    let mut result = String::from_utf8(output.stdout).expect("Got invalid UTF8 from node");
+    let mut result =
+        String::from_utf8(output.stdout).with_context(|| "Got invalid UTF8 from node")?;
     result.push('\n');
     Ok(result)
 }
