@@ -1,7 +1,7 @@
 use super::grammars::{InputGrammar, PrecedenceEntry, Variable, VariableType};
 use super::rules::{Precedence, Rule};
 use anyhow::{anyhow, Result};
-use serde_derive::Deserialize;
+use serde::Deserialize;
 use serde_json::{Map, Value};
 
 #[derive(Deserialize)]
@@ -19,6 +19,7 @@ enum RuleJSON {
     },
     PATTERN {
         value: String,
+        flags: Option<String>,
     },
     SYMBOL {
         name: String,
@@ -143,7 +144,21 @@ fn parse_rule(json: RuleJSON) -> Rule {
         } => Rule::alias(parse_rule(*content), value, named),
         RuleJSON::BLANK => Rule::Blank,
         RuleJSON::STRING { value } => Rule::String(value),
-        RuleJSON::PATTERN { value } => Rule::Pattern(value),
+        RuleJSON::PATTERN { value, flags } => Rule::Pattern(
+            value,
+            flags.map_or(String::new(), |f| {
+                f.chars()
+                    .filter(|c| {
+                        if *c != 'i' {
+                            eprintln!("Warning: unsupported flag {c}");
+                            false
+                        } else {
+                            *c != 'u' // silently ignore unicode flag
+                        }
+                    })
+                    .collect()
+            }),
+        ),
         RuleJSON::SYMBOL { name } => Rule::NamedSymbol(name),
         RuleJSON::CHOICE { members } => Rule::choice(members.into_iter().map(parse_rule).collect()),
         RuleJSON::FIELD { content, name } => Rule::field(name, parse_rule(*content)),
