@@ -49,7 +49,7 @@ pub(super) fn extract_tokens(
         }) = variable.rule
         {
             if i > 0 && extractor.extracted_usage_counts[index] == 1 {
-                let mut lexical_variable = &mut lexical_variables[index];
+                let lexical_variable = &mut lexical_variables[index];
                 lexical_variable.kind = variable.kind;
                 lexical_variable.name = variable.name;
                 symbol_replacer.replacements.insert(i, index);
@@ -209,7 +209,7 @@ impl TokenExtractor {
                 } else {
                     Rule::Metadata {
                         params: params.clone(),
-                        rule: Box::new(self.extract_tokens_in_rule((&rule).clone())),
+                        rule: Box::new(self.extract_tokens_in_rule(&rule)),
                     }
                 }
             }
@@ -320,7 +320,7 @@ mod test {
                 "rule_0",
                 Rule::repeat(Rule::seq(vec![
                     Rule::string("a"),
-                    Rule::pattern("b"),
+                    Rule::pattern("b", ""),
                     Rule::choice(vec![
                         Rule::non_terminal(1),
                         Rule::non_terminal(2),
@@ -331,8 +331,8 @@ mod test {
                     ]),
                 ])),
             ),
-            Variable::named("rule_1", Rule::pattern("e")),
-            Variable::named("rule_2", Rule::pattern("b")),
+            Variable::named("rule_1", Rule::pattern("e", "")),
+            Variable::named("rule_2", Rule::pattern("b", "")),
             Variable::named(
                 "rule_3",
                 Rule::seq(vec![Rule::non_terminal(2), Rule::Blank]),
@@ -378,12 +378,12 @@ mod test {
             lexical_grammar.variables,
             vec![
                 Variable::anonymous("a", Rule::string("a")),
-                Variable::auxiliary("rule_0_token1", Rule::pattern("b")),
+                Variable::auxiliary("rule_0_token1", Rule::pattern("b", "")),
                 Variable::auxiliary(
                     "rule_0_token2",
                     Rule::repeat(Rule::choice(vec![Rule::string("c"), Rule::string("d"),]))
                 ),
-                Variable::named("rule_1", Rule::pattern("e")),
+                Variable::named("rule_1", Rule::pattern("e", "")),
             ]
         );
     }
@@ -411,7 +411,7 @@ mod test {
     fn test_extracting_extra_symbols() {
         let mut grammar = build_grammar(vec![
             Variable::named("rule_0", Rule::string("x")),
-            Variable::named("comment", Rule::pattern("//.*")),
+            Variable::named("comment", Rule::pattern("//.*", "")),
         ]);
         grammar.extra_symbols = vec![Rule::string(" "), Rule::non_terminal(1)];
 
@@ -493,13 +493,7 @@ mod test {
     fn build_grammar(variables: Vec<Variable>) -> InternedGrammar {
         InternedGrammar {
             variables,
-            extra_symbols: Vec::new(),
-            external_tokens: Vec::new(),
-            supertype_symbols: Vec::new(),
-            expected_conflicts: Vec::new(),
-            variables_to_inline: Vec::new(),
-            precedence_orderings: Vec::new(),
-            word_token: None,
+            ..Default::default()
         }
     }
 }
