@@ -1,3 +1,5 @@
+#![doc = include_str!("../README.md")]
+
 pub mod c_lib;
 
 use memchr::memchr;
@@ -16,7 +18,7 @@ use tree_sitter::{
 const MAX_LINE_LEN: usize = 180;
 const CANCELLATION_CHECK_INTERVAL: usize = 100;
 
-/// Contains the data neeeded to compute tags for code written in a
+/// Contains the data needed to compute tags for code written in a
 /// particular language.
 #[derive(Debug)]
 pub struct TagsConfiguration {
@@ -41,7 +43,7 @@ pub struct NamedCapture {
 }
 
 pub struct TagsContext {
-    parser: Parser,
+    pub parser: Parser,
     cursor: QueryCursor,
 }
 
@@ -82,7 +84,6 @@ struct PatternInfo {
 #[derive(Debug)]
 struct LocalDef<'a> {
     name: &'a [u8],
-    value_range: Range<usize>,
 }
 
 #[derive(Debug)]
@@ -135,7 +136,7 @@ impl TagsConfiguration {
         let mut local_scope_capture_index = None;
         let mut local_definition_capture_index = None;
         for (i, name) in query.capture_names().iter().enumerate() {
-            match name.as_str() {
+            match *name {
                 "" => continue,
                 "name" => name_capture_index = Some(i as u32),
                 "ignore" => ignore_capture_index = Some(i as u32),
@@ -253,6 +254,10 @@ impl TagsContext {
         }
     }
 
+    pub fn parser(&mut self) -> &mut Parser {
+        &mut self.parser
+    }
+
     pub fn generate_tags<'a>(
         &'a mut self,
         config: &'a TagsConfiguration,
@@ -349,7 +354,6 @@ where
                             }) {
                                 scope.local_defs.push(LocalDef {
                                     name: &self.source[range.clone()],
-                                    value_range: range,
                                 });
                             }
                         }
